/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// gl_main.c

#include "quakedef.h"

entity_t	r_worldentity;

qboolean	r_cache_thrash;		// compatability

vec3_t		modelorg, r_entorigin;
entity_t	*currententity;

int			r_visframecount;	// bumped when going to a new PVS
int			r_framecount;		// used for dlight push checking

mplane_t	frustum[4];

int			c_brush_polys, c_alias_polys, c_d_lightmaps;


//up to 16 color translated skins
gltexture_t *playertextures[MAX_SCOREBOARD]; // changed to an array of pointers


//
// view origin
//
vec3_t	vup;
vec3_t	vpn;
vec3_t	vright;
vec3_t	r_origin;

float	r_world_matrix[16];
float	r_base_world_matrix[16];

//
// screen size info
//
refdef_t	r_refdef;

mleaf_t		*r_viewleaf, *r_oldviewleaf;

int		d_lightstylevalue[256];	// 8.8 fraction of base light value

float r_fovx, r_fovy;

// Interpolation
void R_RotateForEntity (entity_t *e);
void R_BlendedRotateForEntity (entity_t *e);
void R_SetupAliasBlendedFrame (int frame, aliashdr_t *paliashdr, entity_t* e);
void GL_DrawAliasBlendedFrame (aliashdr_t *paliashdr, int pose1, int pose2, float blend);

cvar_t	r_norefresh = {"r_norefresh","0"};
cvar_t	r_drawentities = {"r_drawentities","1"};
cvar_t	r_drawworld = {"r_drawworld","1"};
cvar_t	r_drawviewmodel = {"r_drawviewmodel","1"};
cvar_t	r_speeds = {"r_speeds","0"};
cvar_t	r_fullbright = {"r_fullbright","0"};
cvar_t	r_wateralpha = {"r_wateralpha","1"};
cvar_t	r_dynamic = {"r_dynamic","1"};
cvar_t	r_novis = {"r_novis","0"};
cvar_t	r_flatlightstyles = {"r_flatlightstyles", "0"};
cvar_t	r_waterwarp = {"r_waterwarp", "1"};
cvar_t	r_clearcolor = {"r_clearcolor", "2"}; // Closest to the original

cvar_t	gl_finish = {"gl_finish","0"};
cvar_t	gl_clear = {"gl_clear","0"};
cvar_t	gl_cull = {"gl_cull","1"};
cvar_t	gl_smoothmodels = {"gl_smoothmodels","1"};
cvar_t	gl_affinemodels = {"gl_affinemodels","0"};
cvar_t	gl_polyblend = {"gl_polyblend","1"};
cvar_t	gl_flashblend = {"gl_flashblend","1"};

/*
=================
R_CullBox
replaced with new function from lordhavoc

Returns true if the box is completely outside the frustum
=================
*/
qboolean R_CullBox (vec3_t emins, vec3_t emaxs)
{
	int i;
	mplane_t *p;
	for (i = 0;i < 4;i++)
	{
		p = frustum + i;
		switch(p->signbits)
		{
			default:
			case 0:
				if (p->normal[0]*emaxs[0] + p->normal[1]*emaxs[1] + p->normal[2]*emaxs[2] < p->dist)
					return true;
				break;
			case 1:
				if (p->normal[0]*emins[0] + p->normal[1]*emaxs[1] + p->normal[2]*emaxs[2] < p->dist)
					return true;
				break;
			case 2:
				if (p->normal[0]*emaxs[0] + p->normal[1]*emins[1] + p->normal[2]*emaxs[2] < p->dist)
					return true;
				break;
			case 3:
				if (p->normal[0]*emins[0] + p->normal[1]*emins[1] + p->normal[2]*emaxs[2] < p->dist)
					return true;
				break;
			case 4:
				if (p->normal[0]*emaxs[0] + p->normal[1]*emaxs[1] + p->normal[2]*emins[2] < p->dist)
					return true;
				break;
			case 5:
				if (p->normal[0]*emins[0] + p->normal[1]*emaxs[1] + p->normal[2]*emins[2] < p->dist)
					return true;
				break;
			case 6:
				if (p->normal[0]*emaxs[0] + p->normal[1]*emins[1] + p->normal[2]*emins[2] < p->dist)
					return true;
				break;
			case 7:
				if (p->normal[0]*emins[0] + p->normal[1]*emins[1] + p->normal[2]*emins[2] < p->dist)
					return true;
				break;
		}
	}
	return false;
}

/*
===============
R_CullModelForEntity

uses correct bounds based on rotation
===============
*/
qboolean R_CullModelForEntity (entity_t *e)
{
	vec3_t mins, maxs;

	if (e->angles[0] || e->angles[2]) // pitch or roll
	{
		VectorAdd (e->origin, e->model->rmins, mins);
		VectorAdd (e->origin, e->model->rmaxs, maxs);
	}
	else if (e->angles[1]) // yaw
	{
		VectorAdd (e->origin, e->model->ymins, mins);
		VectorAdd (e->origin, e->model->ymaxs, maxs);
	}
	else // no rotation
	{
		VectorAdd (e->origin, e->model->mins, mins);
		VectorAdd (e->origin, e->model->maxs, maxs);
	}

	return R_CullBox (mins, maxs);
}


/*
=============================================================

  SPRITE MODELS

=============================================================
*/

/*
================
R_GetSpriteFrame
================
*/
mspriteframe_t *R_GetSpriteFrame (entity_t *currententity)
{
	msprite_t		*psprite;
	mspritegroup_t	*pspritegroup;
	mspriteframe_t	*pspriteframe;
	int				i, numframes, frame;
	float			*pintervals, fullinterval, targettime, time;
	static float	lastmsg = 0;

	psprite = currententity->model->cache.data;
	frame = currententity->frame;

	if (frame >= psprite->numframes || frame < 0)
	{
		if ((IsTimeout (&lastmsg, 2) && (frame == currententity->model->maxframe || frame < 0)) ||
		    frame > currententity->model->maxframe)
		{
			if (frame > currententity->model->maxframe)
				currententity->model->maxframe = frame; // Save peak

			Con_Printf ("R_GetSpriteFrame: no such frame %d (%d frames) in %s\n", frame, psprite->numframes, currententity->model->name);
		}

		frame = 0;
	}

	if (psprite->frames[frame].type == SPR_SINGLE)
	{
		pspriteframe = psprite->frames[frame].frameptr;
	}
	else
	{
		pspritegroup = (mspritegroup_t *)psprite->frames[frame].frameptr;
		pintervals = pspritegroup->intervals;
		numframes = pspritegroup->numframes;
		fullinterval = pintervals[numframes-1];

		time = cl.time + currententity->syncbase;

	// when loading in Mod_LoadSpriteGroup, we guaranteed all interval values
	// are positive, so we don't have to worry about division by 0
		targettime = time - ((int)(time / fullinterval)) * fullinterval;

		for (i=0 ; i<(numframes-1) ; i++)
		{
			if (pintervals[i] > targettime)
				break;
		}

		pspriteframe = pspritegroup->frames[i];
	}

	return pspriteframe;
}


/*
=================
R_DrawSpriteModel 

now supports all orientations
=================
*/
void R_DrawSpriteModel (entity_t *e)
{
	vec3_t			point, v_forward, v_right, v_up;
	msprite_t		*psprite;
	mspriteframe_t	*frame;
	float			*s_up, *s_right;
	float			angle, sr, cr;

	//TODO: frustum cull it?
	frame = R_GetSpriteFrame (e);
	psprite = currententity->model->cache.data;

	switch(psprite->type)
	{
	case SPR_VP_PARALLEL_UPRIGHT: //faces view plane, up is towards the heavens
		v_up[0] = 0;
		v_up[1] = 0;
		v_up[2] = 1;
		s_up = v_up;
		s_right = vright;
		break;
	case SPR_FACING_UPRIGHT: //faces camera origin, up is towards the heavens
		VectorSubtract(currententity->origin, r_origin, v_forward);
		v_forward[2] = 0;
		VectorNormalizeFast(v_forward);
		v_right[0] = v_forward[1];
		v_right[1] = -v_forward[0];
		v_right[2] = 0;
		v_up[0] = 0;
		v_up[1] = 0;
		v_up[2] = 1;
		s_up = v_up;
		s_right = v_right;
		break;
	case SPR_VP_PARALLEL: //faces view plane, up is towards the top of the screen
		s_up = vup;
		s_right = vright;
		break;
	case SPR_ORIENTED: //pitch yaw roll are independent of camera
		AngleVectors (currententity->angles, v_forward, v_right, v_up);
		s_up = v_up;
		s_right = v_right;
		break;
	case SPR_VP_PARALLEL_ORIENTED: //faces view plane, but obeys roll value
		angle = currententity->angles[ROLL] * M_PI_DIV_180;
		sr = sin(angle);
		cr = cos(angle);
		v_right[0] = vright[0] * cr + vup[0] * sr;
		v_right[1] = vright[1] * cr + vup[1] * sr;
		v_right[2] = vright[2] * cr + vup[2] * sr;
		v_up[0] = vright[0] * -sr + vup[0] * cr;
		v_up[1] = vright[1] * -sr + vup[1] * cr;
		v_up[2] = vright[2] * -sr + vup[2] * cr;
		s_up = v_up;
		s_right = v_right;
		break;
	default:
		return;
	}

	// offset decals
	if (psprite->type == SPR_ORIENTED)
	{
		glEnable (GL_POLYGON_OFFSET_FILL);
		glEnable (GL_POLYGON_OFFSET_LINE);
		glPolygonOffset (-1, -10000);
	}

	glColor3f (1,1,1);

	GL_DisableMultitexture (); // selects TEXTURE0
	GL_Bind (frame->gltexture);

	glEnable (GL_ALPHA_TEST);
	glBegin (GL_QUADS);

	glTexCoord2f (0, 1);
	VectorMA (e->origin, frame->down, s_up, point);
	VectorMA (point, frame->left, s_right, point);
	glVertex3fv (point);

	glTexCoord2f (0, 0);
	VectorMA (e->origin, frame->up, s_up, point);
	VectorMA (point, frame->left, s_right, point);
	glVertex3fv (point);

	glTexCoord2f (1, 0);
	VectorMA (e->origin, frame->up, s_up, point);
	VectorMA (point, frame->right, s_right, point);
	glVertex3fv (point);

	glTexCoord2f (1, 1);
	VectorMA (e->origin, frame->down, s_up, point);
	VectorMA (point, frame->right, s_right, point);
	glVertex3fv (point);

	glEnd ();
	glDisable (GL_ALPHA_TEST);

	// offset decals
	if (psprite->type == SPR_ORIENTED)
	{
		glPolygonOffset (0, 0);
		glDisable (GL_POLYGON_OFFSET_LINE);
		glDisable (GL_POLYGON_OFFSET_FILL);
	}
}

/*
=============================================================

  ALIAS MODELS

=============================================================
*/


#define NUMVERTEXNORMALS	162

float	r_avertexnormals[NUMVERTEXNORMALS][3] = {
#include "anorms.h"
};

extern vec3_t	lightcolor; // replaces "float shadelight" for lit support

// precalculated dot products for quantized angles
#define SHADEDOT_QUANT 16
float	r_avertexnormal_dots[SHADEDOT_QUANT][256] =
#include "anorm_dots.h"
;

float	*shadedots = r_avertexnormal_dots[0];

qboolean shading = true; // if false, disable vertex shading for various reasons (fullbright, etc)

float	entalpha;

int	lastposenum;
int	lastposenum0; // Interpolation

/*
=================
R_ChkFrame
=================
*/
int R_ChkFrame (char *function, int frame, aliashdr_t *pahdr, float *plastmsg)
{
	if (frame >= pahdr->numframes || frame < 0)
	{
		if ((IsTimeout (plastmsg, 2) && (frame == currententity->model->maxframe || frame < 0)) ||
		    frame > currententity->model->maxframe)
		{
			if (frame > currententity->model->maxframe)
				currententity->model->maxframe = frame; // Save peak

			Con_DPrintf ("%s: no such frame ", function);

			// Single frame?
			if (pahdr->frames[0].name[0])
				Con_DPrintf ("%d ('%s', %d frames)", frame, pahdr->frames[0].name, pahdr->numframes);
			else
				Con_DPrintf ("group %d (%d groups)", frame, pahdr->numframes);

			Con_DPrintf (" in %s\n", currententity->model->name);
		}

		frame = 0;
	}

	return frame;
}

/*
=================
R_DrawAliasModel

=================
*/
void R_DrawAliasModel (entity_t *e, qboolean isweapon)
{
	int			lnum;
	vec3_t		dist;
	float		add;
	model_t		*clmodel;
	aliashdr_t	*paliashdr;
	int			anim;
	qboolean	isclient = false;
	int			skinnum, client_no;
	gltexture_t	*tx, *fb;
	static float	lastmsg = 0;

	//
	// cull it
	//
	if (R_CullModelForEntity(e))
		return;

	clmodel = e->model;
	client_no = e - cl_entities;

	VectorCopy (e->origin, r_entorigin);
	VectorSubtract (r_origin, r_entorigin, modelorg);

	//
	// locate the proper data
	//
	paliashdr = (aliashdr_t *)Mod_Extradata (e->model);
	skinnum = e->skinnum;

	// r_speeds
	c_alias_polys += paliashdr->numtris;

	// check skin bounds
	if (skinnum >= paliashdr->numskins || skinnum < 0)
	{
		if (IsTimeout (&lastmsg, 2))
			Con_DPrintf ("R_DrawAliasModel: no such skin %d (%d skins) in %s\n", skinnum, paliashdr->numskins, clmodel->name);
		skinnum = 0;
	}

	//
	// transform it
	//
	glPushMatrix ();

	// fenix@io.com: model transform interpolation
	// Don't interpolate the player in chasecam during demo playback
	if (!isweapon && (client_no != cl.viewentity || !chase_active.value || !cls.demoplayback))
		R_BlendedRotateForEntity (e);
	else
		R_RotateForEntity (e);

	glTranslatef (paliashdr->scale_origin[0], paliashdr->scale_origin[1], paliashdr->scale_origin[2]);
	glScalef (paliashdr->scale[0], paliashdr->scale[1], paliashdr->scale[2]);

	//
	// random stuff
	//
	if (gl_smoothmodels.value)
		glShadeModel (GL_SMOOTH);
	if (gl_affinemodels.value)
		glHint (GL_PERSPECTIVE_CORRECTION_HINT, GL_FASTEST);

	shading = true;

	//
	// set up for alpha blending
	//
	entalpha = ENTALPHA_DECODE(e->alpha);

	if (entalpha == 0)
		goto cleanup;

	if (entalpha < 1.0)
	{
		glDepthMask (GL_FALSE);
		glEnable (GL_BLEND);
	}

	//
	// set up lighting
	//
	R_LightPoint (e->origin);

	//add dlights
	for (lnum=0 ; lnum<MAX_DLIGHTS ; lnum++)
	{
		if (cl_dlights[lnum].die >= cl.time)
		{
			VectorSubtract (e->origin, cl_dlights[lnum].origin, dist);
			add = cl_dlights[lnum].radius - Length(dist);

			if (add > 0)
				VectorMA (lightcolor, add, cl_dlights[lnum].color, lightcolor);
		}
	}

	// minimum light value on gun (24)
	if (e == &cl.viewent)
	{
		add = 72.0f - (lightcolor[0] + lightcolor[1] + lightcolor[2]);
		if (add > 0.0f)
		{
			lightcolor[0] += add / 3.0f;
			lightcolor[1] += add / 3.0f;
			lightcolor[2] += add / 3.0f;
		}
	}

	if (client_no >= 1 && client_no <= cl.maxclients)
	{
		isclient = true;

		// minimum light value on players (8)
		add = 24.0f - (lightcolor[0] + lightcolor[1] + lightcolor[2]);
		if (add > 0.0f)
		{
			lightcolor[0] += add / 3.0f;
			lightcolor[1] += add / 3.0f;
			lightcolor[2] += add / 3.0f;
		}
	}

	// clamp lighting so it doesn't overbright as much (96)
	add = 288.0f / (lightcolor[0] + lightcolor[1] + lightcolor[2]);
	if (add < 1.0f)
		VectorScale (lightcolor, add, lightcolor);

	shadedots = r_avertexnormal_dots[((int)(e->angles[1] * (SHADEDOT_QUANT / 360.0))) & (SHADEDOT_QUANT - 1)];
	VectorScale (lightcolor, 1.0f / 192.0f, lightcolor);

	//
	// set up textures
	//
//	GL_DisableMultitexture (); // selects TEXTURE0

	anim = (int)(cl.time*10) & 3;
	tx = paliashdr->gltexture[skinnum][anim];
	fb = paliashdr->fullbright[skinnum][anim];

	// we can't dynamically colormap textures, so they are cached
	// seperately for the players.  Heads are just uncolored.
	if (e->colormap != vid.colormap)
	{
		if (isclient)
			tx = playertextures[client_no - 1];
	}

	//
	// draw it
	//
	if (gl_mtexable && gl_texture_env_combine && gl_texture_env_add && fb) // case 1: everything in one pass
	{
		// Binds normal skin to texture env 0
		GL_DisableMultitexture (); // selects TEXTURE0
		GL_Bind (tx);
		glTexEnvf (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE_EXT);
		glTexEnvf (GL_TEXTURE_ENV, GL_COMBINE_RGB_EXT, GL_MODULATE);
		glTexEnvf (GL_TEXTURE_ENV, GL_SOURCE0_RGB_EXT, GL_TEXTURE);
		glTexEnvf (GL_TEXTURE_ENV, GL_SOURCE1_RGB_EXT, GL_PRIMARY_COLOR_EXT);
		glTexEnvf (GL_TEXTURE_ENV, GL_RGB_SCALE_EXT, 2.0f);

		// Binds fullbright skin to texture env 1
		GL_EnableMultitexture (); // selects TEXTURE1
		GL_Bind (fb);
		glTexEnvf (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_ADD);
		glEnable (GL_BLEND);
		R_SetupAliasBlendedFrame (e->frame, paliashdr, e); // fenix@io.com: model animation interpolation
		glDisable (GL_BLEND);
		GL_DisableMultitexture (); // selects TEXTURE0
		glTexEnvf (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
	}
	else if (gl_texture_env_combine) // case 2: overbright in one pass, then fullbright pass
	{
		// first pass
		GL_Bind (tx);
		glTexEnvf (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE_EXT);
		glTexEnvf (GL_TEXTURE_ENV, GL_COMBINE_RGB_EXT, GL_MODULATE);
		glTexEnvf (GL_TEXTURE_ENV, GL_SOURCE0_RGB_EXT, GL_TEXTURE);
		glTexEnvf (GL_TEXTURE_ENV, GL_SOURCE1_RGB_EXT, GL_PRIMARY_COLOR_EXT);
		glTexEnvf (GL_TEXTURE_ENV, GL_RGB_SCALE_EXT, 2.0f);
		R_SetupAliasBlendedFrame (e->frame, paliashdr, e); // fenix@io.com: model animation interpolation
		glTexEnvf (GL_TEXTURE_ENV, GL_RGB_SCALE_EXT, 1.0f);
		glTexEnvf (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);

		// second pass
		if (fb)
		{
			GL_Bind (fb);
			glTexEnvf (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
			glEnable (GL_BLEND);
			glBlendFunc (GL_ONE, GL_ONE);
			glDepthMask (GL_FALSE);
			shading = false;
			glColor3f (entalpha, entalpha, entalpha);
			R_FogStartAdditive ();
			R_SetupAliasBlendedFrame (e->frame, paliashdr, e); // fenix@io.com: model animation interpolation
			R_FogStopAdditive ();
			glDepthMask (GL_TRUE);
			glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
			glDisable (GL_BLEND);
			glTexEnvf (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
		}
	}
	else // case 3: overbright in two passes, then fullbright pass
	{
		// first pass
		GL_Bind (tx);
		glTexEnvf (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
		R_SetupAliasBlendedFrame (e->frame, paliashdr, e); // fenix@io.com: model animation interpolation

		// second pass -- additive with black fog, to double the object colors but not the fog color
		glEnable (GL_BLEND);
		glBlendFunc (GL_ONE, GL_ONE);
		glDepthMask (GL_FALSE);
		R_FogStartAdditive ();
		R_SetupAliasBlendedFrame (e->frame, paliashdr, e); // fenix@io.com: model animation interpolation
		R_FogStopAdditive ();
		glDepthMask (GL_TRUE);
		glTexEnvf (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
		glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		glDisable (GL_BLEND);

		// third pass
		if (fb)
		{
			GL_Bind (fb);
			glTexEnvf (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
			glEnable (GL_BLEND);
			glBlendFunc (GL_ONE, GL_ONE);
			glDepthMask (GL_FALSE);
			shading = false;
			glColor3f (entalpha, entalpha, entalpha);
			R_FogStartAdditive ();
			R_SetupAliasBlendedFrame (e->frame, paliashdr, e); // fenix@io.com: model animation interpolation
			R_FogStopAdditive ();
			glDepthMask (GL_TRUE);
			glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
			glDisable (GL_BLEND);
			glTexEnvf (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
		}
	}

cleanup:
	glTexEnvf (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
	glHint (GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST); // gl_affinemodels
	glShadeModel (GL_FLAT); // gl_smoothmodels
	glDepthMask (GL_TRUE);
	glDisable (GL_BLEND);
	glColor3f (1, 1, 1);
	glPopMatrix ();

}

//==================================================================================

/*
=============
R_DrawEntitiesOnList
=============
*/
void R_DrawEntitiesOnList (qboolean alphapass)
{
	int		i;

	if (!r_drawentities.value)
		return;

	// sprites are not a special case
	for (i=0 ; i<cl_numvisedicts ; i++)
	{
		if ((i + 1) % 100 == 0)
			S_ExtraUpdateTime (); // don't let sound get messed up if going slow

		currententity = cl_visedicts[i];

		// if alphapass is true, draw only alpha entites this time
		// if alphapass is false, draw only nonalpha entities this time
		if ((ENTALPHA_DECODE(currententity->alpha) < 1 && !alphapass) ||
			(ENTALPHA_DECODE(currententity->alpha) == 1 && alphapass))
			continue;

		// chase_active
		if (currententity == &cl_entities[cl.viewentity])
			currententity->angles[0] *= 0.3;

		switch (currententity->model->type)
		{
			case mod_alias:
				R_DrawAliasModel (currententity, false);
				break;

			case mod_brush:
				R_DrawBrushModel (currententity, false);
				break;

			case mod_sprite:
				R_DrawSpriteModel (currententity);
				break;

			default:
				break;
		}
	}
}

/*
=============
R_DrawWaterEntitiesOnList
=============
*/
void R_DrawWaterEntitiesOnList (qboolean alphapass)
{
	int		i;

	if (!r_drawentities.value)
		return;

	// sprites are not a special case
	for (i=0 ; i<cl_numvisedicts ; i++)
	{
		if ((i + 1) % 100 == 0)
			S_ExtraUpdateTime (); // don't let sound get messed up if going slow

		currententity = cl_visedicts[i];

		// if alphapass is true, draw only alpha entites this time
		// if alphapass is false, draw only nonalpha entities this time
		if ((ENTALPHA_DECODE(currententity->alpha) < 1 && !alphapass) ||
			(ENTALPHA_DECODE(currententity->alpha) == 1 && alphapass))
			continue;

		// chase_active
		if (currententity == &cl_entities[cl.viewentity])
			currententity->angles[0] *= 0.3;

		switch (currententity->model->type)
		{
			case mod_brush:
				R_DrawBrushModel (currententity, true);
				break;

			default:
				break;
		}
	}
}

/*
=============
R_DrawViewModel
=============
*/
void R_DrawViewModel (void)
{
	if (!r_drawviewmodel.value || chase_active.value)
		return;

	if (cl.items & IT_INVISIBILITY || cl.stats[STAT_HEALTH] <= 0)
		return;

	currententity = &cl.viewent;

	if (!currententity->model)
		return;

	// this fixes a crash
	if (currententity->model->type != mod_alias)
		return;

	// Prevent weapon model error
	if (currententity->model->name[0] == '*')
	{
		Con_Warning ("R_DrawViewModel: viewmodel %s invalid\n", currententity->model->name);
		Cvar_Set ("r_drawviewmodel", "0");
		return;
	}


	// hack the depth range to prevent view model from poking into walls
	glDepthRange (0, 0.3);

	// fenix@io.com: model transform interpolation
	R_DrawAliasModel (currententity, true);

	glDepthRange (0, 1);
}


/*
============
R_PolyBlend
============
*/
void R_PolyBlend (void)
{
	if (!gl_polyblend.value)
		return;

	if (!v_blend[3])
		return;

	GL_DisableMultitexture (); // selects TEXTURE0

	glDisable (GL_ALPHA_TEST);
	glDisable (GL_TEXTURE_2D);
	glDisable (GL_DEPTH_TEST);
	glEnable (GL_BLEND);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity ();
	glOrtho (0, 1, 1, 0, -99999, 99999);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity ();

	glColor4fv (v_blend);

	glBegin (GL_QUADS);
	glVertex2f (0,0);
	glVertex2f (1, 0);
	glVertex2f (1, 1);
	glVertex2f (0, 1);
	glEnd ();

	glDisable (GL_BLEND);
	glEnable (GL_DEPTH_TEST);
	glEnable (GL_TEXTURE_2D);
	glEnable (GL_ALPHA_TEST);
}


int SignbitsForPlane (mplane_t *out)
{
	int	bits, j;

	// for fast box on planeside test

	bits = 0;
	for (j=0 ; j<3 ; j++)
	{
		if (out->normal[j] < 0)
			bits |= 1<<j;
	}
	return bits;
}

/*
===============
TurnVector

turn forward towards side on the plane defined by forward and side
if angle = 90, the result will be equal to side
assumes side and forward are perpendicular, and normalized
to turn away from side, use a negative angle
===============
*/
void TurnVector (vec3_t out, vec3_t forward, vec3_t side, float angle)
{
	float scale_forward, scale_side;

	scale_forward = cos (DEG2RAD(angle));
	scale_side = sin (DEG2RAD(angle));

	out[0] = scale_forward*forward[0] + scale_side*side[0];
	out[1] = scale_forward*forward[1] + scale_side*side[1];
	out[2] = scale_forward*forward[2] + scale_side*side[2];
}

/*
===============
R_SetFrustum
===============
*/
void R_SetFrustum (float fovx, float fovy)
{
	int		i;

	TurnVector(frustum[0].normal, vpn, vright, fovx/2 - 90); //left plane
	TurnVector(frustum[1].normal, vpn, vright, 90 - fovx/2); //right plane
	TurnVector(frustum[2].normal, vpn, vup, 90 - fovy/2); //bottom plane
	TurnVector(frustum[3].normal, vpn, vup, fovy/2 - 90); //top plane

	for (i=0 ; i<4 ; i++)
	{
		frustum[i].type = PLANE_ANYZ;
		frustum[i].dist = DotProduct (r_origin, frustum[i].normal); //FIXME: shouldn't this always be zero?
		frustum[i].signbits = SignbitsForPlane (&frustum[i]);
	}
}

/*
=============
R_Clear
=============
*/
void R_Clear (void)
{
	if (gl_clear.value || isIntel) // intel video workaround
		glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	else
		glClear (GL_DEPTH_BUFFER_BIT);
}

/*
===============
R_CheckVariables
===============
*/
void R_CheckVariables (void)
{
	static float oldbright = 0, oldclear = 0, oldgamma = 0;

	if (cls.state == ca_dedicated)
		return;				// stdout only

	if (r_fullbright.value != oldbright)
	{
		oldbright = r_fullbright.value;

		// Refresh lightmaps
		R_BuildLightmaps ();
	}

	if (r_clearcolor.value != oldclear)
	{
		byte *rgb;

		oldclear = r_clearcolor.value;

		// Refresh clearcolor
		rgb = (byte *)(d_8to24table + ((int)r_clearcolor.value & 0xFF));
		glClearColor (rgb[0] / 255.0, rgb[1] / 255.0, rgb[2] / 255.0, 0);
	}

	if (v_gamma.value != oldgamma)
	{
		oldgamma = v_gamma.value;

		// Refresh gamma
		VID_Gamma ();
	}
}

/*
===============
R_SetupFrame
===============
*/
void R_SetupFrame (void)
{
// don't allow cheats in multiplayer
	if (cl.maxclients > 1)
	{
		Cvar_Set ("r_drawentities", "1");
		Cvar_Set ("r_drawworld", "1");
		Cvar_Set ("r_fullbright", "0");
	}

	R_CheckVariables ();
	R_AnimateLight ();

	r_framecount++;

	R_FogSetupFrame ();

// build the transformation matrix for the given view angles
	VectorCopy (r_refdef.vieworg, r_origin);
	AngleVectors (r_refdef.viewangles, vpn, vright, vup);

// current viewleaf
	r_oldviewleaf = r_viewleaf;
	r_viewleaf = Mod_PointInLeaf (r_origin, cl.worldmodel);

	V_SetContentsColor (r_viewleaf->contents);
	V_CalcBlend ();

	r_cache_thrash = false;

	// calculate r_fovx and r_fovy here
	r_fovx = r_refdef.fov_x;
	r_fovy = r_refdef.fov_y;
	if (r_waterwarp.value && (r_viewleaf->contents == CONTENTS_WATER || r_viewleaf->contents == CONTENTS_SLIME || r_viewleaf->contents == CONTENTS_LAVA))
	{
		r_fovx = r_refdef.fov_x + sin(cl.time);
		r_fovy = r_refdef.fov_y - cos(cl.time);
	}

	R_SetFrustum (r_fovx, r_fovy); // use r_fov* vars
	R_MarkSurfaces ();	// create texture chains from PVS (done here so we know if we're in water)
	R_CullSurfaces ();	// do after R_SetFrustum and R_MarkSurfaces
	R_UpdateWarpTextures ();	// do this before R_Clear
	R_Clear ();

	c_brush_polys = 0;
	c_alias_polys = 0;
	c_d_lightmaps = 0;
}

void GL_SetFrustum (float fovx, float fovy)
{
	float xmin, xmax, ymin, ymax;

	xmax = NEARCLIP * tan( fovx * M_PI / 360.0 );
	ymax = NEARCLIP * tan( fovy * M_PI / 360.0 );

	xmin = -xmax;
	ymin = -ymax;

	glFrustum (xmin, xmax, ymin, ymax, NEARCLIP, FARCLIP);
}

/*
=============
R_SetupGL
=============
*/
void R_SetupGL (void)
{
	//
	// set up viewpoint
	//
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity ();
	glViewport (glx + r_refdef.vrect.x,
				gly + glheight - r_refdef.vrect.y - r_refdef.vrect.height,
				r_refdef.vrect.width,
				r_refdef.vrect.height);

	GL_SetFrustum (r_fovx, r_fovy);

//	glCullFace(GL_FRONT);
//	glCullFace(GL_BACK); // glquake used CCW with backwards culling -- let's do it right

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity ();

	glRotatef (-90,  1, 0, 0);	    // put Z going up
	glRotatef (90,  0, 0, 1);	    // put Z going up
	glRotatef (-r_refdef.viewangles[2],  1, 0, 0);
	glRotatef (-r_refdef.viewangles[0],  0, 1, 0);
	glRotatef (-r_refdef.viewangles[1],  0, 0, 1);
	glTranslatef (-r_refdef.vieworg[0],  -r_refdef.vieworg[1],  -r_refdef.vieworg[2]);

	glGetFloatv (GL_MODELVIEW_MATRIX, r_world_matrix);

	//
	// set drawing parms
	//
	if (gl_cull.value)
		glEnable(GL_CULL_FACE);
	else
		glDisable(GL_CULL_FACE);

	glDisable(GL_BLEND);
	glDisable(GL_ALPHA_TEST);
	glEnable(GL_DEPTH_TEST);
}

/*
================
R_RenderView

r_refdef must be set before the first call
================
*/
void R_RenderView (void)
{
	float	time1 = 0, time2;
	float	ms;
	char Str[256]; 

	if (r_norefresh.value)
		return;

	if (!r_worldentity.model || !cl.worldmodel)
		Sys_Error ("R_RenderView: NULL worldmodel");

	if (r_speeds.value)
	{
		glFinish ();
		time1 = Sys_DoubleTime ();
		c_brush_polys = 0;
		c_alias_polys = 0;
		c_d_lightmaps = 0;
	}

	if (gl_finish.value)
		glFinish ();

	// render normal view
	// r_refdef must be set before the first call
	R_SetupFrame ();
	R_SetupGL ();

	S_ExtraUpdateTime ();	// don't let sound get messed up if going slow

	R_FogEnableGFog ();
	R_DrawSky (); // handle worldspawn and bmodels
	R_DrawWorld (); // adds static entities to the list
	R_DrawEntitiesOnList (false); // false means this is the pass for nonalpha entities
	R_DrawWaterEntitiesOnList (false); // special case to draw water nonalpha entities
	R_DrawTextureChainsWater (); // drawn here since they might have transparency
	R_DrawEntitiesOnList (true); // true means this is the pass for alpha entities
	R_DrawWaterEntitiesOnList (true); // special case to draw water alpha entities
	R_RenderDlights ();
	R_DrawParticles ();
	R_FogDisableGFog ();
	R_DrawViewModel ();
	R_PolyBlend ();
	R_BloomBlend (); // bloom on each frame. 

	S_ExtraUpdateTime ();	// don't let sound get messed up if going slow

	if (r_speeds.value)
	{
//		glFinish ();
		time2 = Sys_DoubleTime ();
		ms = 1000 * (time2 - time1);
		sprintf (Str, "%5.1f ms %4i wpoly %4i epoly %4i lmaps\n", ms, c_brush_polys, c_alias_polys, c_d_lightmaps); 
		Con_Printf (Str);
	}
}

/*
=============
R_RotateForEntity
=============
*/
void R_RotateForEntity (entity_t *e)
{
	glTranslatef (e->origin[0], e->origin[1], e->origin[2]);

	glRotatef (e->angles[1], 0, 0, 1);
	glRotatef (-e->angles[0], 0, 1, 0);
	glRotatef (e->angles[2], 1, 0, 0);
}

/*
=============
GL_DrawAliasBlendedFrame

fenix@io.com: model animation interpolation
=============
*/
void GL_DrawAliasBlendedFrame (aliashdr_t *paliashdr, int pose1, int pose2, float blend)
{
	float 	vertcolor[4]; // replaces "float l" for lit support
	trivertx_t	*verts1;
	trivertx_t	*verts2;
	int			*commands;
	int			count;
	float		u,v;
	float		iblend;

	lastposenum0 = pose1;
	lastposenum  = pose2;
	verts1  = (trivertx_t *)((byte *)paliashdr + paliashdr->posedata);
	verts2  = verts1;

	verts1 += pose1 * paliashdr->poseverts;
	verts2 += pose2 * paliashdr->poseverts;

	iblend = 1.0f - blend;

	commands = (int *)((byte *)paliashdr + paliashdr->commands);

	vertcolor[3] = entalpha; // never changes, so there's no need to put this inside the loop

	while (1)
	{
		// get the vertex count and primitive type
		count = *commands++;

		if (!count) 
			break; // done

		if (count < 0)
		{
			count = -count;
			glBegin (GL_TRIANGLE_FAN);
		}
		else
			glBegin (GL_TRIANGLE_STRIP);

		do
		{
			// texture coordinates come from the draw list
			u = ((float *)commands)[0];
			v = ((float *)commands)[1];

			if (mtexenabled)
			{
				qglMultiTexCoord2f (TEXTURE0, u, v);
				qglMultiTexCoord2f (TEXTURE1, u, v);
			}
			else
			{
				glTexCoord2f (u, v);
			}

			commands += 2;

			if (shading)
			{
				// lit support
				if (r_fullbright.value || !cl.worldmodel->lightdata)
					vertcolor[0] = vertcolor[1] = vertcolor[2] = 1.0;
				else
				{
					vertcolor[0] = (shadedots[verts1->lightnormalindex]*iblend + shadedots[verts2->lightnormalindex]*blend) * lightcolor[0];
					vertcolor[1] = (shadedots[verts1->lightnormalindex]*iblend + shadedots[verts2->lightnormalindex]*blend) * lightcolor[1];
					vertcolor[2] = (shadedots[verts1->lightnormalindex]*iblend + shadedots[verts2->lightnormalindex]*blend) * lightcolor[2];
				}
				glColor4fv (vertcolor);
			}

			glVertex3f (verts1->v[0]*iblend + verts2->v[0]*blend,
						verts1->v[1]*iblend + verts2->v[1]*blend,
						verts1->v[2]*iblend + verts2->v[2]*blend);
			verts1++;
			verts2++;
		}
		while (--count);

		glEnd ();
	}
}

static float TimePassed (float lasttime)
{
	float timepassed = cl.time - lasttime;

	if (timepassed < 0)
	{
		if (timepassed > -0.01)
			timepassed = 0; // This happens in demos
/*		else
		{
			Con_Warning ("TimePassed: timepassed = %f\n", timepassed);
		} */
	}
	
	return timepassed;
}

/*
=================
R_SetupAliasBlendedFrame

fenix@io.com: model animation interpolation
=================
*/
void R_SetupAliasBlendedFrame (int frame, aliashdr_t *paliashdr, entity_t* e)
{
	int	     pose;
	int	     numposes;
	float    blend, timepassed;
	static float lastmsg = 0;

	frame = R_ChkFrame ("R_SetupAliasBlendedFrame", frame, paliashdr, &lastmsg);

	pose = paliashdr->frames[frame].firstpose;
	numposes = paliashdr->frames[frame].numposes;

	if (numposes > 1)
	{
		int firstpose = pose;

		pose = numposes * currententity->syncbase; // Hack to make flames unsynchronized
		e->frame_interval = paliashdr->frames[frame].interval;
		pose += (int)(cl.time / e->frame_interval) % numposes;
		pose = firstpose + pose % numposes;
	}
	else
	{
		/* One tenth of a second is a good for most Quake animations.
		If the nextthink is longer then the animation is usually meant to pause
		(e.g. check out the shambler magic animation in shambler.qc).  If its
		shorter then things will still be smoothed partly, and the jumps will be
		less noticable because of the shorter time.  So, this is probably a good
		assumption. */
		e->frame_interval = 0.1;
	}

	timepassed = TimePassed (e->frame_start_time);

	if (e->pose2 != pose || timepassed < 0)
	{
		e->frame_start_time = cl.time;
		e->pose1 = e->pose2;
		e->pose2 = pose;
		blend = 0;
	}
	else
		blend = timepassed / e->frame_interval;

	// wierd things start happening if blend passes 1
	if (cl.paused || blend > 1)
		blend = 1;
/*
	if (e->pose1 >= paliashdr->numposes || e->pose2 >= paliashdr->numposes)
	{
		Con_Warning ("R_SetupAliasBlendedFrame: pose1 (%d) or pose2 (%d) is >= numposes (%d) in %s\n", e->pose1, e->pose2, paliashdr->numposes, e->model->name);
	}
*/
	// Prevent crash later in GL_DrawAliasBlendedFrame, probably
	// caused by entity morphing e.g. from monster to gibbed head
	if (e->pose1 >= paliashdr->numposes)
		e->pose1 = 0; 
	
	if (e->pose2 >= paliashdr->numposes)
		e->pose2 = 0; // I haven't seen this happen, but just in case ...
		
	GL_DrawAliasBlendedFrame (paliashdr, e->pose1, e->pose2, blend);
}

/*
=============
R_BlendedRotateForEntity

fenix@io.com: model transform interpolation
=============
*/
void R_BlendedRotateForEntity (entity_t *e)
{
	float 	timepassed;
	float 	blend;
	vec3_t	d;
	int	i;

	// positional interpolation

	timepassed = TimePassed (e->translate_start_time);

	if (e->translate_start_time == 0 || timepassed > 1 || timepassed < 0)
	{
		timepassed = 0;
		e->translate_start_time = cl.time;
		VectorCopy (e->origin, e->origin1);
		VectorCopy (e->origin, e->origin2);
	}

	if (!VectorCompare (e->origin, e->origin2))
	{
		e->translate_start_time = cl.time;
		VectorCopy (e->origin2, e->origin1);
		VectorCopy (e->origin,  e->origin2);
		blend = 0;
	}
	else
	{
		blend =  timepassed / 0.1;

		if (cl.paused || blend > 1)
			blend = 1;
	}

	VectorSubtract (e->origin2, e->origin1, d);

	glTranslatef (e->origin1[0] + (blend * d[0]),
		      e->origin1[1] + (blend * d[1]),
		      e->origin1[2] + (blend * d[2]));

	// orientation interpolation (Euler angles, yuck!)

	timepassed = TimePassed (e->rotate_start_time);

	if (e->rotate_start_time == 0 || timepassed > 1 || timepassed < 0)
	{
		timepassed = 0;
		e->rotate_start_time = cl.time;
		VectorCopy (e->angles, e->angles1);
		VectorCopy (e->angles, e->angles2);
	}

	if (!VectorCompare (e->angles, e->angles2))
	{
		e->rotate_start_time = cl.time;
		VectorCopy (e->angles2, e->angles1);
		VectorCopy (e->angles,  e->angles2);
		blend = 0;
	}
	else
	{
		blend =  timepassed / 0.1;

		if (cl.paused || blend > 1)
			blend = 1;
	}

	VectorSubtract (e->angles2, e->angles1, d);

	// always interpolate along the shortest path
	for (i = 0; i < 3; i++)
	{
		if (d[i] > 180)
			d[i] -= 360;
		else if (d[i] < -180)
			d[i] += 360;
	}

	glRotatef (e->angles1[1] + ( blend * d[1]),  0, 0, 1);
	glRotatef (-e->angles1[0] + (-blend * d[0]),  0, 1, 0);
	glRotatef ( e->angles1[2] + ( blend * d[2]),  1, 0, 0);
}
